#!/usr/bin/perl
#
# Copyright (C) 2020  BaseALT /basealt.ru/
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#

use strict;
use Getopt::Long;
use Pod::Usage;

my $outFile = "";
my $logDir = "";
my $noarch = 0;
my $help = 0;

GetOptions ("o|out=s" => \$outFile,
            "l|logdir=s"   => \$logDir,
            "n|noarch" => \$noarch,
            "help|h" => \$help) or pod2usage(2);

pod2usage(1) if $help;

my %state = ();
my %laststate = ();

sub push_state ($$) {
    my $nvr = shift;
    my $arch = shift;

    if ($nvr) {
        if ($noarch) {
            if ($laststate{$nvr} && %{$laststate{$nvr}}) {
                foreach my $arch (keys %{$laststate{$nvr}}) {
                    my %h = %{$laststate{$nvr}{$arch}};
                    $h{'arch'} = '';
                    push @{$state{$nvr}}, \%h;
                }
                foreach my $arch (keys %{$laststate{$nvr}}) {
                    delete $laststate{$nvr}{$arch};
                }
            }
        } else {
            if ($nvr && $laststate{$nvr}{$arch} && %{$laststate{$nvr}{$arch}}) {
                my %h = %{$laststate{$nvr}{$arch}};
                push @{$state{$nvr}{$arch}}, \%h;
            }
        }
    }
}

while (<>) {
    my $logln = "$_";
    my $nvr = "";
    my $arch = "";

    if (/^---([0-9]+)\t(\S+)\t(\S*)\t(\S+)\t(\S*)\t(\S*)\t(\S*)\t(\S*)\t(\S*)/) {
        %{$laststate{$2}{$3}} = ("try" => $1,
                                 "class" => $4,
                                 "state" => $5,
                                 "substate" => $6,
                                 "estate" => $7,
                                 "mem" => $8,
                                 "vfyerr" => $9,
                                 "log" => $laststate{$2}{$3}{"log"});
        next;
    }

    if (/^(\[([^\s\]]+)\]:)?ERROR: File is not accessible: (.+\/)?(.+)\.src\.rpm/) {
        push_state ($4, "");
        %{$laststate{$4}{""}} = ("class" => "REBUILD",
                                 "state" => "NOTFOUND",
                                 "log" => "$logln");
        next;
    }

    next unless s/^(\[([^\s\]]+)\]:)?\[(#[0-9]+\s+)?([^\s\]]+)(\s+([^\s\]]+))?\]\s+//;

    $nvr = $4;
    $arch = $6;

    if (/^Start rebuilding/) {
        push_state ($nvr, $arch);
        %{$laststate{$nvr}{$arch}} = ("class" => "REBUILD");
    } elsif (/^Importing the package to git.../) {
        push_state ($nvr, $arch);
        %{$laststate{$nvr}{$arch}} = ("class" => "CLEAN");
    } elsif (/^Creating build environment.../) {
        push_state ($nvr, $arch);
        %{$laststate{$nvr}{$arch}} = ("class" => "SRPM");
    } elsif (/^Skip: no need to rebuild\.$/) {
        push_state ($nvr, $arch);
        %{$laststate{$nvr}{$arch}} = ("class" => "REBUILD",
                                      "state" => "ALRDY");
    } elsif (/No changes\.$/) {
        push_state ($nvr, $arch);
        %{$laststate{$nvr}{$arch}} = ("class" => "CLEAN",
                                      "state" => "NOCHGS");
        next;
    }

    next unless $laststate{$nvr}{$arch}{"class"};

    if ($laststate{$nvr}{$arch}{"class"} eq "REBUILD") {
        if (/Architecture\s+is\s+(not\s+included|excluded)/) {
            $laststate{$nvr}{$arch}{"substate"} = "EXCL";
        } elsif (/hsh-install: Failed to calculate package file list/) {
            if (($laststate{$nvr}{$arch}{"substate"} ne "NOSPACE") && ($laststate{$nvr}{$arch}{"substate"} ne "NOINODE")) {
                $laststate{$nvr}{$arch}{"substate"} = "BRKNDEP";
            }
        } elsif (/Bad exit status from.*(%check)/) {
            $laststate{$nvr}{$arch}{"substate"} = "CHKERR";
        } elsif (/No space left on device/ || /needs .*B on the .* filesystem/) {
            $laststate{$nvr}{$arch}{"substate"} = "NOSPACE";
        } elsif (/needs .* inodes on the .* filesystem/) {
            $laststate{$nvr}{$arch}{"substate"} = "NOINODE";
        } elsif (/^ERROR: Verification error( \((\S+)% of errors\))?/) {
            $laststate{$nvr}{$arch}{"substate"} = "VFYERR";
            $laststate{$nvr}{$arch}{"vfyerr"} = $2;
        } elsif (/^Verification OK( \((\S+)% of errors\))?/) {
            $laststate{$nvr}{$arch}{"substate"} = "VFYOK";
            $laststate{$nvr}{$arch}{"vfyerr"} = $2;
        } elsif (/^([0-9]+)\s+total$/) {
            $laststate{$nvr}{$arch}{"mem"} = $1;
        } elsif (/^ERROR: Unsuccessfully finished rebuilding/) {
            $laststate{$nvr}{$arch}{"state"} = "BLDERR";
        } elsif (/^Successfully finished rebuilding/) {
            $laststate{$nvr}{$arch}{"state"} = "BLDOK";
        } elsif (/^ERROR: Log processing error/) {
            $laststate{$nvr}{$arch}{"state"} = "LOGERR";
        } elsif (/^Done processing log\.$/) {
            $laststate{$nvr}{$arch}{"state"} = "LOGOK";
        } elsif (/^ERROR: Unable to export RPMs/) {
            $laststate{$nvr}{$arch}{"estate"} = "EXPERR";
        } elsif (/^Copying the resulting RPMs to/) {
            $laststate{$nvr}{$arch}{"estate"} = "CPOK";
        } elsif (/^ERROR: Failed to copy RPMs to/) {
            $laststate{$nvr}{$arch}{"estate"} = "CPERR";
        } elsif (/^Done\.$/) {
            if ($laststate{$nvr}{$arch}{"state"} eq "BLDOK") {
                $laststate{$nvr}{$arch}{"state"} = "LOGOK";
            }
        }
    } elsif ($laststate{$nvr}{$arch}{"class"} eq "CLEAN") {
        if (/^No changes/) {
            $laststate{$nvr}{$arch}{"state"} = "NOCHGS";
        } elsif (/^ERROR: Unable to gear-srpmimport/) {
            $laststate{$nvr}{$arch}{"state"} = "IMPERR";
        }  elsif (/^ERROR: Failed to truncate files/) {
            $laststate{$nvr}{$arch}{"state"} = "TRNCERR";
        }  elsif (/^ERROR: Failed to write \S+\.tar/) {
            $laststate{$nvr}{$arch}{"state"} = "GEARERR";
        } elsif (/^Done writing \S+\.tar/) {
            $laststate{$nvr}{$arch}{"state"} = "TAROK";
        }
    } elsif ($laststate{$nvr}{$arch}{"class"} eq "SRPM") {
        if (/hsh-install: Failed to calculate package file list/) {
            if (($laststate{$nvr}{$arch}{"substate"} ne "NOSPACE") && ($laststate{$nvr}{$arch}{"substate"} ne "NOINODE")) {
                $laststate{$nvr}{$arch}{"substate"} = "BRKNDEP";
            }
        } elsif (/No space left on device/ || /needs .*B on the .* filesystem/) {
            $laststate{$nvr}{$arch}{"substate"} = "NOSPACE";
        } elsif (/needs .* inodes on the .* filesystem/) {
            $laststate{$nvr}{$arch}{"substate"} = "NOINODE";
        } elsif (/^ERROR: Failed to create build environment/) {
            $laststate{$nvr}{$arch}{"state"} = "ENVERR";
        } elsif (/^ERROR: Failed to build the source package/) {
            $laststate{$nvr}{$arch}{"state"} = "BLDERR";
        } elsif (/^Done building source package/) {
            $laststate{$nvr}{$arch}{"state"} = "BLDOK";
        } elsif (/^ERROR: Failed to copy new SRPM package/) {
            $laststate{$nvr}{$arch}{"estate"} = "CPERR";
        } elsif (/^Copy modified SRPM package to/) {
            $laststate{$nvr}{$arch}{"estate"} = "CPOK";
        }
    }

    $laststate{$nvr}{$arch}{"log"} .= "$logln";
}

foreach my $nvr (keys %laststate) {
    foreach my $arch (keys %{$laststate{$nvr}}) {
        push_state ($nvr, $arch);
    }
}

if ($outFile) {
    open (OUT, ">$outFile") or die "ERROR: Unable to open $outFile: $!";
} else {
    open (OUT, ">&=STDOUT") or die "ERROR: Unable to redirect to stdout: $!";
}

sub print_state($$;$) {
    my $nvr = shift;
    my $pkgstates = shift;
    my $arch = shift unless $noarch;

    return unless $#$pkgstates >= 0;

    my $pkgtry = 0;
    foreach (@{$pkgstates}) {
        if ($_->{'try'} > $pkgtry) {
            $pkgtry = $_->{'try'};
        }
    }
    $pkgtry += 1;

    my %pkgstate = %{$$pkgstates[$#$pkgstates]};

    if ($pkgstate{'try'}) {
        $pkgtry = $pkgstate{'try'};
    }

    if ($noarch) {
        $arch = $pkgstate{'arch'};
    }

    print OUT "$pkgstate{'log'}";

    print OUT "---$pkgtry";
    print OUT "\t$nvr";
    print OUT "\t$arch";
    print OUT "\t$pkgstate{'class'}";
    print OUT "\t$pkgstate{'state'}";
    print OUT "\t$pkgstate{'substate'}";
    print OUT "\t$pkgstate{'estate'}";
    print OUT "\t$pkgstate{'mem'}";
    print OUT "\t$pkgstate{'vfyerr'}";
    print OUT "\n";
}

if ($noarch) {
    foreach my $nvr (sort keys %state) {
        my @pkgstates =  @{$state{$nvr}};
        print_state ($nvr, \@pkgstates);
    }
} else {
    foreach my $nvr (sort keys %state) {
        foreach my $arch (sort keys %{$state{$nvr}}) {
            my @pkgstates =  @{$state{$nvr}{$arch}};
            print_state ($nvr, \@pkgstates, $arch);
        }
    }
}

END {
    if (fileno OUT) {
        close (OUT) or die "ERROR: Unable to close " . ($outFile ? $outFile : "standard output") . ": $!";
    }
}

__END__

=head1 NAME

import-logs.pl - Reads hsh-separate-sources(1), srpm-cleanup(1) and
status logs and output the resulting status log.

=head1 SYNOPSIS

import-logs.pl [options] [LOGFILE...]

=head1 OPTIONS

=over 4

=item B<-o|--out> STATUSLOG

Output the status log to STATUSLOG instead of standard output.

=item B<-l|--logdir> LOGDIR

Sets the path to the build logs directory.

=item B<-h|--help>

Print a brief help message and exit.

=back

=head1 DESCRIPTION

This program is for internal usage by srpm-cleanup's Makefile
(see make-srpm-cleanup(7)).

=cut
